#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <gtk/gtk.h>

#include "../include/string.h"
#include "../include/disk.h"

#include "guiutils.h"
#include "cdialog.h"
#include "editclist.h"

#include "editor.h"
#include "editorobj.h"
#include "editortex.h"
#include "editorlist.h"
#include "editorcheck.h"
#include "config.h"


static gboolean EditorCheckTextureNameExists(
	editclist_struct *eclist, const gchar *s
);
static gint EditorDoCheckObjectsIterate(
        editor_struct *e, gboolean verbose, gboolean yes_to_all,
	gint *warnings, gint *errors,
        gint on, gpointer obj
);
gint EditorDoCheckObjects(
	editor_struct *e, gboolean verbose,
	gint *warnings, gint *errors
);
gint EditorDoCheckHeader(
	editor_struct *e, gboolean verbose,
	gint *warnings, gint *errors
);
gint EditorDoCheck(editor_struct *e, gboolean verbose);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)     (((x) < 0) ? ((x) * -1) : (x))
#define IS_STRING_EMPTY(s)      (((s) != NULL) ? (*(s) == '\0') : TRUE)

#define RADTODEG(r)     ((r) * 180 / PI)
#define DEGTORAD(d)     ((d) * PI / 180)

#define STATUS_MESSAGE(s)	{ if(verbose) StatusBarMessage(sb, s, TRUE); }


/*
 *	Returns TRUE if s is in the editclist's list of textures.
 */
static gboolean EditorCheckTextureNameExists(
        editclist_struct *eclist, const gchar *s
)
{
	gint row;
	gchar *s2;
	guint8 spacing;
	GdkPixmap *pixmap;
	GdkBitmap *mask;
	GtkCList *clist = (GtkCList *)((eclist != NULL) ?
	    EDITCLIST_CLIST(eclist) : NULL
	);
	if((clist == NULL) || (s == NULL))
	    return(FALSE);

	for(row = 0; row < clist->rows; row++)
	{
	    s2 = NULL;
	    switch((gint)gtk_clist_get_cell_type(clist, row, 0))
	    {
	      case GTK_CELL_TEXT:
		gtk_clist_get_text(clist, row, 0, &s2);
		break;
	      case GTK_CELL_PIXTEXT:
		gtk_clist_get_pixtext(
		    clist, row, 0, &s2, &spacing, &pixmap, &mask
		);
		break;
	    }
	    if(s2 == NULL)
		continue;

	    if(!g_strcasecmp(s2, s))
		return(TRUE);
	}

	return(FALSE);
}


/*
 *	Called by EditorDoCheckObjects().
 */
static gint EditorDoCheckObjectsIterate(
	editor_struct *e, gboolean verbose, gboolean yes_to_all,
	gint *warnings, gint *errors,
	gint on, gpointer obj
)
{
	gint status = 0;
	const gchar *name;
        GtkWidget *toplevel;
        status_bar_struct *sb;
	editor_obj_struct *o = EDITOR_OBJ(obj);
	if(o == NULL)
	    return(2);

	toplevel = EDITOR_TOPLEVEL(e);
	sb = EDITOR_STATUS_BAR(e);
	name = o->name;

#define RAISE_STATUS(i)                         		\
{ if((status == 0) || (i < status)) { status = i; }		\
  if((i == 1) && (errors != NULL)) *errors = *errors + 1;	\
  if((i == 2) && (warnings != NULL)) *warnings = *warnings + 1;	\
}
#define NOTIFY_WARNING(s)                       \
{ if(verbose && !yes_to_all) {                  \
  gint response;                                \
  gchar *title;					\
  if(IS_STRING_EMPTY(name))			\
   title = g_strdup_printf(			\
    "Object #%i Warning", on			\
   );						\
  else						\
   title = g_strdup_printf(                     \
    "Object \"%s\" Warning", name		\
   );						\
  CDialogSetTransientFor(toplevel);             \
  response = CDialogGetResponse(                \
   title,					\
   s, NULL, CDIALOG_ICON_WARNING,               \
   CDIALOG_BTNFLAG_OK |                         \
   CDIALOG_BTNFLAG_YES_TO_ALL,                  \
   CDIALOG_BTNFLAG_OK                           \
  );                                            \
  g_free(title);				\
  g_free(s);                                    \
  CDialogSetTransientFor(NULL);                 \
  if(response == CDIALOG_RESPONSE_YES_TO_ALL)   \
   yes_to_all = TRUE;                           \
} }
#define NOTIFY_ERROR(s)                         \
{ if(verbose && !yes_to_all) {                  \
  gint response;                                \
  gchar *title;                                 \
  if(IS_STRING_EMPTY(name))                     \
   title = g_strdup_printf(                     \
    "Object #%i Error", on			\
   );                                           \
  else                                          \
   title = g_strdup_printf(                     \
    "Object \"%s\" Error", name			\
   );                                           \
  CDialogSetTransientFor(toplevel);             \
  response = CDialogGetResponse(                \
   title,					\
   s, NULL, CDIALOG_ICON_ERROR,                 \
   CDIALOG_BTNFLAG_OK |                         \
   CDIALOG_BTNFLAG_YES_TO_ALL,                  \
   CDIALOG_BTNFLAG_OK                           \
  );                                            \
  g_free(title);				\
  g_free(s);                                    \
  CDialogSetTransientFor(NULL);                 \
  if(response == CDIALOG_RESPONSE_YES_TO_ALL)   \
   yes_to_all = TRUE;                           \
} }

	/* Name. */
	if(!IS_STRING_EMPTY(name))
	{
	    if(strchr(name, ' '))
	    {
		NOTIFY_ERROR(STRDUP(
"Name may not contain spaces"
		));
		RAISE_STATUS(1);
	    }
            else if(strchr(name, '\t'))
            {
                NOTIFY_ERROR(STRDUP(
"Name may not contain spaces"
                ));
                RAISE_STATUS(1);
            }
            else if(strchr(name, '='))
            {
                NOTIFY_ERROR(STRDUP(
"Name may not contain \"=\" characters"
                ));
                RAISE_STATUS(1);
            }
	    /* Check if special name is being used. */
	    else if(EditorIsObjectNameSpecial(name))
	    {
                NOTIFY_ERROR(g_strdup_printf(
"The name \"%s\" is reserved for special use,\n\
please choose a different name",
		    name
                ));
                RAISE_STATUS(1);
	    }

	    /* Check if another object has this same exact name. */
	    if(e->obj_clist != NULL)
	    {
		gint i;
		editor_obj_struct *o2;
		GtkCList *clist = (GtkCList *)e->obj_clist;

		/* Iterate through all objects. */
		for(i = 0; i < clist->rows; i++)
		{
		    /* Skip this object. */
		    if(i == on)
			continue;

		    o2 = EDITOR_OBJ(gtk_clist_get_row_data(
			clist, i
		    ));
		    if(o2 == NULL)
			continue;

		    /* This object has a name? */
		    if(!IS_STRING_EMPTY(o2->name))
		    {
			if(!g_strcasecmp(o2->name, name))
			{
			    NOTIFY_ERROR(g_strdup_printf(
"Object \"%s\" (#%i) has the same name as this object.",
				o2->name, i
			    ));
			    RAISE_STATUS(1);
			}
		    }
		}	/* Iterate through all objects. */
	    }


	}
	/* Range. */
	if(o->range < 0.0f)
	{
	    NOTIFY_WARNING(STRDUP(
"Range cannot be negative, you must specify a positive or 0 value\n\
for the range.\n\
\n\
Specifying 0 means to use the default range."
	    ));
	    RAISE_STATUS(2);
	}
        /* Range far. */
        if(o->range_far < 0.0f)
        {
            NOTIFY_WARNING(STRDUP(
"Far range cannot be negative, you must specify a positive or 0 value\n\
for the far range.\n\
\n\
Specifying 0 means to use the default far range."
            ));
            RAISE_STATUS(2);
        }
	/* Model file. */
	if(!IS_STRING_EMPTY(o->model_file))
	{
	    const gchar *filename = o->model_file;
	    gchar *path;
	    struct stat stat_buf;

	    if(ISPATHABSOLUTE(filename))
		path = STRDUP(filename);
	    else
#ifdef __MSW__
		path = STRDUP(filename);
#else
		path = g_strdup_printf(
		    "%s%c%s",
		    SE_DEF_SAR_DATA_DIR, DIR_DELIMINATOR, filename
		);
#endif
	    if(stat(path, &stat_buf))
	    {
		NOTIFY_WARNING(g_strdup_printf(
"Unable to find model file:\n\n    %s",
		    path
		));
		RAISE_STATUS(2);
	    }
#ifdef S_ISDIR
	    else if(S_ISDIR(stat_buf.st_mode))
	    {
                NOTIFY_WARNING(g_strdup_printf(
"The model file:\n\n    %s\n\nIs a directory",
                    path
                ));
		RAISE_STATUS(2);
	    }
#endif
	    g_free(path);
	}
	/* Reference object name. */
	if(!IS_STRING_EMPTY(o->ref_obj_name))
	{
	    const gchar *ref_obj_name = o->ref_obj_name;

	    /* Check if object name is a special name. */
	    if(EditorIsObjectNameSpecial(ref_obj_name))
	    {


	    }
	    else
	    {
	        gint ref_obj_num = EditorObjListMatchByName(e, ref_obj_name);
		if(ref_obj_num < 0)
		{
		    NOTIFY_WARNING(g_strdup_printf(
"Unable to find object referenced by the name \"%s\"",
			ref_obj_name
		    ));
		    RAISE_STATUS(2);
		}
		else if(ref_obj_num == on)
		{
		    NOTIFY_WARNING(g_strdup_printf(
"Object referenced by the name \"%s\" is this object itself",
			ref_obj_name
		    ));
		    RAISE_STATUS(2);
		}
            }
	}

	/* Helipad. */
	if(o->type == SAR_OBJ_TYPE_HELIPAD)
	{
	    const gchar *helipad_label = o->helipad_label;
	    if(!IS_STRING_EMPTY(helipad_label))
	    {
		if(strchr(helipad_label, ' '))
                {
                    NOTIFY_ERROR(STRDUP(
"Helipad label may not contain spaces (suggest using \"_\" instead)"
                    ));
                    RAISE_STATUS(1);
                }
                if(strchr(helipad_label, '\t'))
                {
                    NOTIFY_ERROR(STRDUP(
"Helipad label may not contain spaces (suggest using \"_\" instead)"
                    ));
                    RAISE_STATUS(1);
                }
                if(strchr(helipad_label, '='))
                {
                    NOTIFY_ERROR(STRDUP(
"Helipad label may not contain \"=\" characters"
                    ));
                    RAISE_STATUS(1);
                }
            }
	}
        /* Runway. */
        else if(o->type == SAR_OBJ_TYPE_RUNWAY)
        {
            const gchar *label = o->runway_north_label;
            if(!IS_STRING_EMPTY(label))
            {
                if(strchr(label, ' '))
                {
                    NOTIFY_ERROR(STRDUP(
"Runway label may not contain spaces (suggest using \"_\" instead)"
                    ));
                    RAISE_STATUS(1);
                }
                if(strchr(label, '\t'))
                {
                    NOTIFY_ERROR(STRDUP(
"Runway label may not contain spaces (suggest using \"_\" instead)"
                    ));
                    RAISE_STATUS(1);
                }
                if(strchr(label, '='))
                {
                    NOTIFY_ERROR(STRDUP(
"Runway label may not contain \"=\" characters"
                    ));
                    RAISE_STATUS(1);
                }
            }
	    label = o->runway_south_label;
            if(!IS_STRING_EMPTY(label))
            {
                if(strchr(label, ' '))
                {
                    NOTIFY_ERROR(STRDUP(
"Runway label may not contain spaces (suggest using \"_\" instead)"
                    ));
                    RAISE_STATUS(1);
                }
                if(strchr(label, '\t'))
                {
                    NOTIFY_ERROR(STRDUP(
"Runway label may not contain spaces (suggest using \"_\" instead)"
                    ));
                    RAISE_STATUS(1);
                }
                if(strchr(label, '='))
                {
                    NOTIFY_ERROR(STRDUP(
"Runway label may not contain \"=\" characters"
                    ));
                    RAISE_STATUS(1);
                }
            }
        }



#undef NOTIFY_ERROR
#undef NOTIFY_WARNING
#undef RAISE_STATUS

	return(status);
}

/*
 *      Checks if all objects in the scenery are okay.
 *
 *      If verbose is TRUE then the user will be notified of any
 *      errors.
 *
 *      Return values (lowest positive number is most severe):
 *
 *      0       Success
 *      1       Error
 *      2       Warning
 *      3       Systems error
 *      4       User abort
 *      5       Unsupported/other
 */
gint EditorDoCheckObjects(
	editor_struct *e, gboolean verbose,
	gint *warnings, gint *errors
)
{
        gboolean yes_to_all = FALSE;
        gint row, status = 0;
        GtkWidget *toplevel;
	GtkCList *clist;
        status_bar_struct *sb;


        if(e == NULL)
            return(1);

        toplevel = EDITOR_TOPLEVEL(e);
        sb = EDITOR_STATUS_BAR(e);
	clist = (GtkCList *)e->obj_clist;
	if(clist == NULL)
	    return(1);

        EditorStopReset(e, TRUE);
        EditorSetBusy(e, TRUE);

	/* Need to apply changes to the object so that we have the
	 * latest values.
	 */
	EditorObjListDoApplyValues(e, NULL);

        STATUS_MESSAGE("Checking objects...");

#define RAISE_STATUS(i)                                         \
{ if((status == 0) || (i < status)) { status = i; }             \
  if((i == 1) && (errors != NULL)) *errors = *errors + 1;       \
  if((i == 2) && (warnings != NULL)) *warnings = *warnings + 1; \
}
#define NOTIFY_WARNING(s)                       \
{ if(verbose && !yes_to_all) {                  \
  gint response;                                \
  CDialogSetTransientFor(toplevel);             \
  response = CDialogGetResponse(                \
   "Scenery Check Object Warning",              \
   s, NULL, CDIALOG_ICON_WARNING,               \
   CDIALOG_BTNFLAG_OK |                         \
   CDIALOG_BTNFLAG_YES_TO_ALL,                  \
   CDIALOG_BTNFLAG_OK                           \
  );                                            \
  g_free(s);                                    \
  CDialogSetTransientFor(NULL);                 \
  if(response == CDIALOG_RESPONSE_YES_TO_ALL)   \
   yes_to_all = TRUE;                           \
} }
#define NOTIFY_ERROR(s)                         \
{ if(verbose && !yes_to_all) {                  \
  gint response;                                \
  CDialogSetTransientFor(toplevel);             \
  response = CDialogGetResponse(                \
   "Scenery Check Object Error",                \
   s, NULL, CDIALOG_ICON_ERROR,                 \
   CDIALOG_BTNFLAG_OK |                         \
   CDIALOG_BTNFLAG_YES_TO_ALL,                  \
   CDIALOG_BTNFLAG_OK                           \
  );                                            \
  g_free(s);                                    \
  CDialogSetTransientFor(NULL);                 \
  if(response == CDIALOG_RESPONSE_YES_TO_ALL)   \
   yes_to_all = TRUE;                           \
} }

	/* Iterate through objects. */
	for(row = 0; row < clist->rows; row++)
	{
	    gint status2;
	    gchar *buf;

	    buf = g_strdup_printf(
		"Checking object #%i...", row
	    );
	    STATUS_MESSAGE(buf);
	    g_free(buf);
            StatusBarProgress(sb, (gfloat)(row + 1) / (gfloat)clist->rows, TRUE);

	    status2 = EditorDoCheckObjectsIterate(
		e, verbose, yes_to_all,
		warnings, errors,
		row, gtk_clist_get_row_data(clist, row)
	    );
	    if(status2)
		RAISE_STATUS(status2);

	    if(EDITOR_STOP_COUNT(e) > 0)
		break;
	}

        EditorStopReset(e, FALSE);
        EditorSetBusy(e, FALSE);

        STATUS_MESSAGE("Objects check done");
	StatusBarProgress(sb, 0.0f, TRUE);

#undef NOTIFY_ERROR
#undef NOTIFY_WARNING
#undef RAISE_STATUS

        return(status);
}

/*
 *      Checks if all scenery header data is okay.
 *
 *      If verbose is TRUE then the user will be notified of any
 *      errors.
 *
 *      Return values (lowest positive number is most severe):
 *
 *      0       Success
 *      1       Error
 *      2       Warning
 *      3       Systems error
 *      4       User abort
 *      5       Unsupported/other
 */
gint EditorDoCheckHeader(
	editor_struct *e, gboolean verbose,
	gint *warnings, gint *errors
)
{
	gboolean yes_to_all = FALSE;
	const gchar *s;
	gint status = 0;
	GtkAdjustment *adj;
	GtkWidget *toplevel;
	GtkEntry *entry;
	GtkSpinButton *spin;
	editor_tex_struct *t;
	status_bar_struct *sb;


        if(e == NULL)
            return(1);

	toplevel = EDITOR_TOPLEVEL(e);
	sb = EDITOR_STATUS_BAR(e);

        EditorStopReset(e, TRUE);
	EditorSetBusy(e, TRUE);

        STATUS_MESSAGE("Checking header...");

#define RAISE_STATUS(i)                                         \
{ if((status == 0) || (i < status)) { status = i; }             \
  if((i == 1) && (errors != NULL)) *errors = *errors + 1;       \
  if((i == 2) && (warnings != NULL)) *warnings = *warnings + 1; \
}
#define NOTIFY_WARNING(s)			\
{ if(verbose && !yes_to_all) {			\
  gint response;				\
  CDialogSetTransientFor(toplevel);		\
  response = CDialogGetResponse(		\
   "Scenery Check Header Warning",		\
   s, NULL, CDIALOG_ICON_WARNING,		\
   CDIALOG_BTNFLAG_OK |				\
   CDIALOG_BTNFLAG_YES_TO_ALL,			\
   CDIALOG_BTNFLAG_OK				\
  );						\
  g_free(s);					\
  CDialogSetTransientFor(NULL);			\
  if(response == CDIALOG_RESPONSE_YES_TO_ALL)	\
   yes_to_all = TRUE;				\
} }
#define NOTIFY_ERROR(s)				\
{ if(verbose && !yes_to_all) {			\
  gint response;				\
  CDialogSetTransientFor(toplevel);		\
  response = CDialogGetResponse(		\
   "Scenery Check Header Error",		\
   s, NULL, CDIALOG_ICON_ERROR,			\
   CDIALOG_BTNFLAG_OK |				\
   CDIALOG_BTNFLAG_YES_TO_ALL,			\
   CDIALOG_BTNFLAG_OK				\
  );						\
  g_free(s);                                    \
  CDialogSetTransientFor(NULL);                 \
  if(response == CDIALOG_RESPONSE_YES_TO_ALL)   \
   yes_to_all = TRUE;                           \
} }


	/* Planetary radius. */
	spin = (GtkSpinButton *)e->hdr_planetary_radius_spin;
	adj = (spin != NULL) ? spin->adjustment : NULL;
	if(adj != NULL)
	{
	    if(adj->value < 1000000.0f)
	    {
		NOTIFY_WARNING(g_strdup_printf(
"Planetary radius %.1f is too small, suggested value is 6371000.0",
		    adj->value
		));
		RAISE_STATUS(2);
	    }
	}

        /* Ground tile. */
        spin = (GtkSpinButton *)e->hdr_ground_tile_width_spin;
        adj = (spin != NULL) ? spin->adjustment : NULL;
        if(adj != NULL)
        {
            if(adj->value <= 1.0f)
	    {
                NOTIFY_WARNING(g_strdup_printf(
"Ground tile X length %.0f is too small",
                    adj->value
                ));
		RAISE_STATUS(2);
	    }
        }
        spin = (GtkSpinButton *)e->hdr_ground_tile_height_spin;
        adj = (spin != NULL) ? spin->adjustment : NULL;
        if(adj != NULL)
        {
            if(adj->value <= 1.0f)
	    {
                NOTIFY_WARNING(g_strdup_printf(
"Ground tile Y length %.0f is too small",
                    adj->value
                ));
		RAISE_STATUS(2);
	    }
        }
	/* Ground tile close range. */
        spin = (GtkSpinButton *)e->hdr_ground_tile_close_range_spin;
        adj = (spin != NULL) ? spin->adjustment : NULL;
        if(adj != NULL)
        {
            if(adj->value <= 1.0f)
	    {
                NOTIFY_WARNING(g_strdup_printf(
"Ground textured tile range %.0f is too small",
                    adj->value
                ));
		RAISE_STATUS(2);
	    }
        }
	/* Ground texture name. */
	entry = (GtkEntry *)e->hdr_ground_texture_entry;
	s = (entry != NULL) ? gtk_entry_get_text(entry) : NULL;
	if(!IS_STRING_EMPTY(s))
	{
	    if(!EditorCheckTextureNameExists(
		e->hdr_textures_eclist, s
	    ))
	    {
		NOTIFY_WARNING(g_strdup_printf(
"Ground texture name \"%s\" was not found in\n\
the textures list.",
		    s
                ));
                RAISE_STATUS(2);
	    }
	}


	/* Map file. */
	t = e->hdr_map_tex;
	if(t != NULL)
	{
            spin = (GtkSpinButton *)e->hdr_map_width_spin;
            adj = (spin != NULL) ? spin->adjustment : NULL;
            if(adj != NULL)
            {
                if(adj->value <= 1.0f)
                {
                    NOTIFY_WARNING(g_strdup_printf(
"Ground map X length %.2f is too small",
                        adj->value
                    ));
                    RAISE_STATUS(2);
                }
            }
            spin = (GtkSpinButton *)e->hdr_map_height_spin;
            adj = (spin != NULL) ? spin->adjustment : NULL;
            if(adj != NULL)
            {
                if(adj->value <= 1.0f)
                {
                    NOTIFY_WARNING(g_strdup_printf(
"Ground map Y length %.2f is too small",
                        adj->value
                    ));
                    RAISE_STATUS(2);
                }
            }

	}


	EditorStopReset(e, FALSE);
        EditorSetBusy(e, FALSE);

	STATUS_MESSAGE("Header check done");

#undef NOTIFY_ERROR
#undef NOTIFY_WARNING
#undef RAISE_STATUS

        return(status);
}


/*
 *	Checks if all scenery data is okay.
 *
 *	If verbose is TRUE then the user will be notified of any
 *	errors.
 *
 *	Return values (lowest positive number is most severe):
 *
 *	0	Success
 *	1	Error
 *	2	Warning
 *	3	Systems error
 *	4	User abort
 *	5	Unsupported/other
 */
gint EditorDoCheck(editor_struct *e, gboolean verbose)
{
	gint status = 0, i;
	gint warnings = 0, errors = 0;
	GtkWidget *toplevel;
	status_bar_struct *sb;


	if(e == NULL)
	    return(1);

	toplevel = EDITOR_TOPLEVEL(e);
        sb = EDITOR_STATUS_BAR(e);

        STATUS_MESSAGE("Checking scenery...");

	/* Check header. */
	i = EditorDoCheckHeader(e, verbose, &warnings, &errors);
	status = MAX(status, i);

	/* Check objects. */
	i = EditorDoCheckObjects(e, verbose, &warnings, &errors);
	status = MAX(status, 1);

	/* Need to print results? */
	if(verbose)
	{
	    /* Any warnings or errors? */
	    if((warnings > 0) || (errors > 0))
	    {
		gchar *buf = g_strdup_printf(
		    "Scenery check done (%i warning%s, %i error%s)",
		    warnings, (warnings != 1) ? "s" : "",
		    errors, (errors != 1) ? "s" : ""
		);
		CDialogSetTransientFor(toplevel);
		CDialogGetResponse(
		    "Scenery Check Results",
		    buf,
		    NULL,
		    CDIALOG_ICON_INFO,
		    CDIALOG_BTNFLAG_OK,
		    CDIALOG_BTNFLAG_OK
		);
		CDialogSetTransientFor(NULL);
		STATUS_MESSAGE(buf);
		g_free(buf);
	    }
	    else
	    {
		STATUS_MESSAGE("Scenery check done");
	    }
	}

	return(status);
}
