#include <gtk/gtk.h>
#include "statusbar.h"


status_bar_struct *StatusBarNew(GtkWidget *parent);
void StatusBarUpdateMenus(status_bar_struct *sb);
void StatusBarMap(status_bar_struct *sb);
void StatusBarUnmap(status_bar_struct *sb);
void StatusBarDelete(status_bar_struct *sb);

void StatusBarMessage(
        status_bar_struct *sb, const gchar *message,
        gboolean allow_gtk_iteration
);
void StatusBarProgress(
        status_bar_struct *sb, gfloat percent,
        gboolean allow_gtk_iteration
);


#define STATUS_BAR_HEIGHT	26


/*
 *	Creates a new status bar.
 *
 *	Given parent is assumed to be a GtkBox.
 */
status_bar_struct *StatusBarNew(GtkWidget *parent)
{
	gint border_minor = 2;
	GtkAdjustment *adj;
        GtkWidget *w, *parent2, *parent3;
	status_bar_struct *sb = STATUS_BAR(g_malloc0(
	    sizeof(status_bar_struct)
	));
	if(sb == NULL)
	    return(NULL);

        /* Toplevel frame. */
        sb->toplevel = w = gtk_frame_new(NULL);
        gtk_widget_set_usize(w, -1, STATUS_BAR_HEIGHT);
        gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_OUT);
        gtk_container_border_width(GTK_CONTAINER(w), 0);
	if(GTK_IS_BOX(parent))
	    gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	else
	    gtk_container_add(GTK_CONTAINER(parent), w);
        parent2 = w;

        /* Table in main frame. */
        w = gtk_table_new(1, 2, FALSE);
        gtk_container_add(GTK_CONTAINER(parent2), w);
        gtk_widget_show(w);
        parent2 = w;


        /* Progress bar. */
        adj = (GtkAdjustment *)gtk_adjustment_new(
	    0.0f, 1.0f, 100.0f, 0.0f, 0.0f, 0.0f
	);
        sb->progress_bar = w = gtk_progress_bar_new_with_adjustment(adj);
        gtk_widget_set_usize(w, 100, 20);
        gtk_progress_bar_set_orientation(
            GTK_PROGRESS_BAR(w), GTK_PROGRESS_LEFT_TO_RIGHT
        );
        gtk_progress_bar_set_bar_style(
            GTK_PROGRESS_BAR(w), GTK_PROGRESS_CONTINUOUS
        );
        gtk_progress_set_activity_mode(
            GTK_PROGRESS(w), FALSE
        );
        gtk_table_attach(
            GTK_TABLE(parent2), w,
            0, 1, 0, 1,
            0,
            GTK_SHRINK,
            0, 0
        );
        gtk_widget_show(w);

        /* Frame for label. */
        w = gtk_frame_new(NULL);
        gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
        gtk_container_border_width(GTK_CONTAINER(w), 1);
        gtk_table_attach(
            GTK_TABLE(parent2), w,
            1, 2, 0, 1,
            GTK_SHRINK | GTK_EXPAND | GTK_FILL,
            GTK_FILL,
            0, 0
        );
        gtk_widget_show(w);
        parent3 = w;
	/* Hbox to align label. */
        w = gtk_hbox_new(FALSE, 0);
        gtk_container_add(GTK_CONTAINER(parent3), w);
        gtk_widget_show(w);
        parent3 = w;
	/* Label. */
        sb->label = w = gtk_label_new("");
        gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, border_minor);
        gtk_widget_show(w);

	return(sb);
}

/*
 *	Updates menus and widgets on the status bar.
 */
void StatusBarUpdateMenus(status_bar_struct *sb)
{
	if(sb == NULL)
	    return;


}

/*
 *	Maps the status bar.
 */
void StatusBarMap(status_bar_struct *sb)
{
        GtkWidget *w;

        if(sb == NULL)
            return;

        w = sb->toplevel;
        if(w != NULL)
            gtk_widget_show(w);
}

/*
 *	Unmaps the status bar.
 */
void StatusBarUnmap(status_bar_struct *sb)
{
	GtkWidget *w;

	if(sb == NULL)
	    return;

	w = sb->toplevel;
	if(w != NULL)
	    gtk_widget_hide(w);
}

/*
 *	Destroys the given status bar.
 */
void StatusBarDelete(status_bar_struct *sb)
{
	GtkWidget **w;


	if(sb == NULL)
	    return;

#define DO_DESTROY_WIDGET       \
{ if(*w != NULL) { gtk_widget_destroy(*w); *w = NULL; } }
	w = &sb->progress_bar;
	DO_DESTROY_WIDGET
	w = &sb->label;
	DO_DESTROY_WIDGET
	w = &sb->toplevel;
	DO_DESTROY_WIDGET
#undef DO_DESTROY_WIDGET

	g_free(sb);
}


/*
 *	Updates the message on the status bar.
 */
void StatusBarMessage(
        status_bar_struct *sb, const gchar *message,
        gboolean allow_gtk_iteration
)
{
        GtkWidget *w;


	if(sb == NULL)
	    return;

        w = sb->label;
        if(w == NULL)
            return;

        gtk_label_set_text(
            GTK_LABEL(w),
            (message != NULL) ? message : ""
        );

        /* Iterate through GTK main function until all events are handled
         * so that we ensure this label text is updated.
         */
        while((gtk_events_pending() > 0) && allow_gtk_iteration)
            gtk_main_iteration();
}

/*
 *	Updates the progress on the status bar.
 *
 *	If value is between 0.0 to 1.0 then normal finite progress
 *	is performed, otherwise if value is negative then unknown time
 *	progress is performed.
 */
void StatusBarProgress(
        status_bar_struct *sb, gfloat value,
        gboolean allow_gtk_iteration
)
{
        GtkWidget *w;
	GtkProgress *progress;


        if(sb == NULL)
            return;

        w = sb->progress_bar;
        if(w == NULL)
            return;
	else
	    progress = GTK_PROGRESS(w);


        /* Negative? Implies just do activity. */
        if(value < 0.0f)
        {
            GtkAdjustment *adj = progress->adjustment;


	    /* Get new value based on unknown position. */
            value = gtk_progress_get_value(progress) + 1.0f;
            if(value > adj->upper)
                value = adj->lower;

	    /* Update progress widget for `activity mode' (unknown limit)
	     * and set new value.
	     */
            gtk_progress_set_activity_mode(progress, TRUE);
            gtk_progress_set_show_text(progress, FALSE);
            gtk_progress_set_value(progress, value);

	    /* Reset last progress position to -1.0, indicating no last
	     * position since in unknown time limit we don't keep track of
	     * the last position.
	     */
	    sb->progress_pos_last = -1.0;

            /* Iterate through GTK main function until all events are
             * handled so that we ensure this progress bar setting is
             * updated.
             */
            while((gtk_events_pending() > 0) && allow_gtk_iteration)
                gtk_main_iteration();

            return;
        }

	/* Clip value to 0.0, 1.0. */
        if(value > 1.0f)
            value = 1.0f;

	/* Reset last progress position if it is greater than the given
	 * value, implying the progress wraped back to the beginning.
	 */
	if(sb->progress_pos_last > value)
	    sb->progress_pos_last = -1.0f;

	/* Check if percent did not sufficiently change (use
	 * whole percent values).
	 */
	if((sb->progress_pos_last > 0.0f) &&
           ((gint)((value - sb->progress_pos_last) * 100) <= 0)
	)
	    return;

	/* Update progress. */
        gtk_progress_set_activity_mode(progress, FALSE);
        gtk_progress_set_format_string(
	    progress, "%p%%"
        );
        gtk_progress_set_show_text(
            progress, (value > 0.0f) ? TRUE : FALSE
        );
	gtk_progress_bar_update(GTK_PROGRESS_BAR(w), value);

        /* Record last position. */
        sb->progress_pos_last = value;

	/* Iterate through GTK main function until all events are
	 * handled so that we ensure this progress bar setting is
	 * updated.
	 */
	while((gtk_events_pending() > 0) && allow_gtk_iteration)
	    gtk_main_iteration();
}
